#include "DNSResolver.h"
DNSResolver::DNSResolver(int32 max)
:fcachesem(NULL),fprogresssem(NULL),fadditionsem(NULL)
,finprogresslist(NULL),fhostnamelist(NULL),fhostinfolist(NULL)
,fmax(max),finit(false)
{
  if (fmax < 1)
  fmax = 1;
}
DNSResolver::DNSResolver(DNSResolver&)
:fcachesem(NULL),fprogresssem(NULL),fadditionsem(NULL)
,finprogresslist(NULL),fhostnamelist(NULL),fhostinfolist(NULL)
,finit(false)
{

}
    
DNSResolver::~DNSResolver()
{
  Uninit();
}
DNSResolver& DNSResolver::operator = (DNSResolver&)
{
  return *this;
}
    
bool DNSResolver::Init()
{
  if (!InitOK())
  {
    finprogresslist = new InProgressList();
    if (finprogresslist == NULL)
    goto common_error;
    
    fhostnamelist = new HostNameList();
    if (fhostnamelist == NULL)
    goto inprogresslist_error;
    
    fhostinfolist = new HostInfoList();
    if (fhostinfolist == NULL)
    goto hostnamelist_error;
      
    fcachesem = new MReaderSWriter();
    if (fcachesem == NULL)
    goto hostinfolist_error;
    
    if (!fcachesem->InitLock())
    goto cachesem_error;
    
    fprogresssem = new SReaderSWriter();
    if (fprogresssem == NULL)
    goto cachesem_error;
    
    if (!fprogresssem->InitLock())
    goto progresssem_error;
    
    fadditionsem = new SReaderSWriter();    
    if (fadditionsem == NULL)
    goto progresssem_error;

    if (!fadditionsem->InitLock())
    goto additionsem_error;

    finit = true;
  }
  return InitOK();
  
  
  additionsem_error:
  delete fadditionsem;

  progresssem_error:
  delete fprogresssem;
  
  cachesem_error:
  delete fcachesem;
  
  hostinfolist_error:
  delete fhostinfolist;

  hostnamelist_error:
  delete fhostnamelist;
  
  inprogresslist_error:
  delete finprogresslist;
  
  common_error:
  return false; 
}
bool DNSResolver::InitOK()const
{
  return finit;
}

void DNSResolver::Uninit()
{
  if (InitOK())
  {
    delete fhostinfolist;
    delete fhostnamelist;
    delete finprogresslist;
    delete fcachesem;
    delete fprogresssem;
    finit = false;
  }
}

status_t DNSResolver::Lookup(const char *hostname,struct sockaddr_in *addr,bool usecache = true)
{
  
  if (usecache)
  return LookupUseCache(hostname,addr);
  struct hostent *hostresult;
  
  if ((hostresult = gethostbyname(hostname)) == NULL)
  return h_errno;

  addr->sin_addr.s_addr = *((uint32 *) hostresult->h_addr);
  return B_OK;

}
status_t DNSResolver::LookupUseCache(const char *hostname,struct sockaddr_in *addr)
{
  HostInfoItem *hostinfoitem;
  InProgressItem *inprogressitem;
  struct hostent *hostresult; 
  int hostnameindex;
  status_t inprogressresult;

  /* ============================================= */
  /* Look up in Hostname list                      */
  /* ============================================= */

  // Lock CacheSem R
  fcachesem->AcquireReader();
  // Finding Hostname item in Cache 
  hostnameindex = fhostnamelist->FindHostName(hostname);
  // Hostname is in cache list ?
  if (hostnameindex >= 0)
  {
    // Get HostNameItem
    hostinfoitem = (HostInfoItem *) fhostnamelist->ItemAt(hostnameindex);
    // Update TimeStamp of HostInfoItem // LRU replacement policy
    hostinfoitem->UpdateTimeStamp();
    // copy the IP address 
    addr->sin_addr.s_addr = hostinfoitem->IPAddress();
    // Unlock CacheSem R
    fcachesem->ReleaseReader();  
    return B_OK;
  }
  else
  // Unlock CacheSem R
  fcachesem->ReleaseReader();

  /* ============================================= */
  /* End of Look up in Hostname list               */
  /* ============================================= */
   


  // HostName is not in cache list
  /* ============================================= */
  /* Look up in inProgress list                    */
  /* ============================================= */
  
  // ==================================== 
  // Lock additional sem
  fadditionsem->AcquireWriter();
  // Unlock additional sem  
  fadditionsem->ReleaseWriter();
  // ==================================== 
  
  // Lock Progress Sem
  fprogresssem->AcquireWriter(); 
  // Check if there is an InProgress Item has a hostname same as user input hostname
  inprogressitem = finprogresslist->FindItemByHostName(hostname);
  
  /* ============================================= */
  /* End of Look up in inProgress list             */
  /* ============================================= */


  // Yes, there is another thread currently evaluate hostinfo of this hostname
  if (inprogressitem != NULL)
  {
    // UW(P)
    fprogresssem->ReleaseWriter();

    /* ================================= */
    /* Manipulate the object             */
    /* ================================= */
    // L(O)
    inprogressitem->AcquireItemSem();
    // register itself to waiting queue for this in progress lookup item
    inprogressitem->WaitCount(inprogressitem->WaitCount() + 1);
    // U(O)
    inprogressitem->ReleaseItemSem();
    /* ================================= */
    /* End of Manipulate the object      */
    /* ================================= */


    /* ================================= */
    /* Wait for result                   */
    /* ================================= */
    // wait until this in progress item is released with the result
    // right now the object holding no other lock
    inprogressitem->AcquireWaitSem();
    /* ================================= */
    /* End of Wait for result            */
    /* ================================= */
    
    
    /* ============================================= */
    /* copy the result of this inprogressitem        */
    /* ============================================= */
    inprogressresult = inprogressitem->Result();
    // if the return result is OK with NO_ERROR
    if (inprogressresult == B_OK)
    {
      // copy the IP Address result 
      addr->sin_addr.s_addr = inprogressitem->IPAddress();
    }
    /* ============================================= */
    /* End of copy the result						 */
    /* ============================================= */
    
    /* ============================================= */
    /* Signal other thread                           */
    /* ============================================= */
    // L(A)
    fadditionsem->AcquireWriter();
    // L(O)
    inprogressitem->AcquireItemSem();
    if (inprogressitem->WaitCount() > 0)
    {
      // there is at least one thread waiting for this result
      // WaitCount--
      inprogressitem->WaitCount(inprogressitem->WaitCount() - 1);
      // U(O)
      inprogressitem->ReleaseItemSem();
      // U(A)
      fadditionsem->ReleaseWriter();
      // Signal()
      inprogressitem->ReleaseWaitSem();
    }
    else
    {
      // there is no thread waiting on this inprogress data object

      // U(O)
      inprogressitem->ReleaseItemSem();
      
      /* ==================================================== */
      /* Remove inProgress Item From inProgressList           */
      /* ==================================================== */
      // L(P)
      fprogresssem->AcquireWriter();
      // Remove inprogress item from the inprogress list
      finprogresslist->Remove(inprogressitem);
      // U(P)
      fprogresssem->ReleaseWriter();
      /* ==================================================== */
      /* End of Remove inProgress Item From inProgressList    */
      /* ==================================================== */
      
      // U(A)
      fadditionsem->ReleaseWriter();
      
      // delete inprogress object 
      delete inprogressitem;
    }      
    return inprogressresult;
  }
  
  /* ==================================== */
  /* Create InProgressItem Object         */
  /* ==================================== */
  if ((inprogressitem = new InProgressItem()) == NULL)
  {
    // UnLock Progress Sem
    fprogresssem->ReleaseWriter();
    return B_NO_MEMORY;  
  }
  
  /* ==================================== */
  /* Init InProgressItem Semaphore        */
  /* ==================================== */
  if (inprogressitem->InitSem() != B_OK)
  {
    delete inprogressitem;
    // U(P)
    fprogresssem->ReleaseWriter();
    return B_NO_MORE_SEMS;  
  }    
  
  /* ==================================== */
  /* Initialize NewInProgress Item        */
  /* ==================================== */
  // initialize waitcount = 0
  inprogressitem->WaitCount(0);
  // Initialize Hostname 
  inprogressitem->HostName(hostname);
  /* ==================================== */
  /* End of Initialize NewInProgress Item */
  /* ==================================== */

  // Add NewInProgress Item into InProgress List
  finprogresslist->AddItem(inprogressitem);
  // U(P)
  fprogresssem->ReleaseWriter();

  /* ========================================= */
  /* LOOK UP IP Address of host name 		   */
  /* ========================================= */
  if ((hostresult = gethostbyname(hostname)) == NULL)
  {
    inprogressresult = h_errno;
    inprogressitem->Result(inprogressresult);
  }
  else
  {
    inprogressresult = B_OK;
    addr->sin_addr.s_addr = *((uint32 *) hostresult->h_addr);
    inprogressitem->Result(inprogressresult);    
    inprogressitem->IPAddress(addr->sin_addr.s_addr);
  }
  /* ========================================= */
  /* End of LOOK UP IP Address of host name    */
  /* ========================================= */

  
  /* ========================================= */
  /* Create new HostInfoItem                   */
  /* ========================================= */
  hostinfoitem = new HostInfoItem(hostname,inprogressitem->IPAddress());
  if (hostinfoitem == NULL)
  return inprogressresult;
  
  /* ========================================= */
  /* Update TimeStamp of HostInfoItem          */
  hostinfoitem->UpdateTimeStamp();
  /* ========================================= */
  
  // LW(H)
  fcachesem->AcquireWriter();
  /* ========================================= */
  /* Resize Cache to fmaxsize - 1              */
  /* ========================================= */
  fhostinfolist->Sort();
  while ((fhostinfolist->CountItems() >= fmax) && (fhostinfolist->CountItems() > 0))
  {
    int32 del_index = fhostinfolist->CountItems() - 1;
    HostInfoItem *del_item = (HostInfoItem *) fhostinfolist->ItemAt(del_index);
    fhostinfolist->RemoveItem(del_index);
    fhostnamelist->RemoveItem(del_item);
    delete del_item;     
  } 
  /* ========================================= */
  /* End of Resize Cache                       */
  /* ========================================= */
    
  /* ============================================ */
  /* Add hostinfo item into hostnamelist          */
  /* ============================================ */
  // LW(H)
  // Add Item to into HostInfoList
  fhostinfolist->AddItem(hostinfoitem);

  // Add Item to into HostNameList
  fhostnamelist->AddItem(hostinfoitem);
  // Sort Host Name List
  fhostnamelist->Sort();

  // UW(H)    
  fcachesem->ReleaseWriter();
  /* ============================================ */
  /* End of Add hostinfo item into hostnamelist   */
  /* ============================================ */
    
  /* ============================================= */
  /* Signal other thread                           */
  /* ============================================= */
  // L(A)
  fadditionsem->AcquireWriter();
  // L(O)
  inprogressitem->AcquireItemSem();
  if (inprogressitem->WaitCount() > 0)
  {
    // there is at least one thread waiting for this result
    // WaitCount--
    inprogressitem->WaitCount(inprogressitem->WaitCount() - 1);
    // U(O)
    inprogressitem->ReleaseItemSem();
    // U(A)
    fadditionsem->ReleaseWriter();
    // Signal()
    inprogressitem->ReleaseWaitSem();
  }
  else
  {
    // there is no thread waiting on this inprogress data object

    // U(O)
    inprogressitem->ReleaseItemSem();
      
    /* ==================================================== */
    /* Remove inProgress Item From inProgressList           */
    /* ==================================================== */
    // L(P)
    fprogresssem->AcquireWriter();
    // Remove inprogress item from the inprogress list
    finprogresslist->Remove(inprogressitem);
    // U(P)
    fprogresssem->ReleaseWriter();
    /* ==================================================== */
    /* End of Remove inProgress Item From inProgressList    */
    /* ==================================================== */
      
    // U(A)
    fadditionsem->ReleaseWriter();
    
    // delete inprogress object 
    delete inprogressitem;
  }      
  return inprogressresult;  
}
void DNSResolver::CacheSize(int32 max)
{
  if (max < 1)
  fmax = 1;
  else
  fmax = max;
   
  fhostinfolist->Sort();
  fcachesem->AcquireWriter();
  while ((fhostinfolist->CountItems() > fmax) && (fhostinfolist->CountItems() > 0))
  {
    int32 del_index = fhostinfolist->CountItems() - 1;
    HostInfoItem *del_item = (HostInfoItem *) fhostinfolist->ItemAt(del_index);
    fhostinfolist->RemoveItem(del_index);
    fhostnamelist->RemoveItem(del_item);
    delete del_item;     
  }
  fcachesem->ReleaseWriter(); 
}
int32 DNSResolver::CacheSize()const
{
  return fmax;
}